

/****************************************************************************

Script:		Scale Move Tool
Version:	0.02
Author:		Joel Hewitt
Email:		joelhewitt@gmail.com

Description:
Allows you to move objects using the scale tool in a similar way that
sub-objects such as verts can be moved

Installation:
1.	Unzip the package to your max root directory
2.	Go to Customize - Customize User Interface... and add the macro to a
	toolbar / quad menu etc. The macro is in the catagory: JH_Tools

Usage:
1.	Select the objects you want to 'Scale Move'
2.	Press the "Start Scale Move" button
3.	Points will be created for each object in the selection and given an 
	instance of the current object's position and rotation controllers.
4.	Move/Scale/Rotate the Points
5.	When you are done, press the "Stop Scale Move" button or simply close
	the dialog - the points will be deleted and any previous hierarchy 
	will be restored to the objects.
	
Extra Options:
-The "Ignore Linked Objects" and "Don't Affect Children" checkboxs in the
 "Hierarchy Options" Group do as you would expect them to.
-Group Options are fairly self explanitory.
-Point Skewing Mode allows you to pick 2 master points and have all other
 points between them move in proportion to their original positions.
-The buttons in the "Quick Alignment Tools" group will align any selected 
 objects pivots in the chosen axis. Does NOT work on Bipeds, Bones or 
 Daylight Systems.

Additional Notes:
-Currently I have not tested the script for animation purposes. Feel free to 
 do so and email me the results.
-If you choose to undo the script to a stage where the points still exist, it
 should be noted that at that time, the objects will no longer have parents
 OR children if the "Don't Affect Children" button is checked and so the 
 previous hierarchy, if any, will need to be restored  manually.
Bugs:
0.01 -No known bugs
0.02 -Fixed a bug with the quickAlign_fn function

Version Update:
0.01 -Initial Release
0.02 -Added the sortLinkedObjs_fn function to work with the quickAlign_fn 
	  function, fixing a problem with aligning linked objects.

 
****************************************************************************/

(
	--Globals	
	global scaleMoveTool
	
	--Private Globals
	local thePoints = #()
	local theObjects = undefined
	local theObjectsParent
	local theObjectsChildren
	local selFilter 
	
	local pOne = undefined
	local pTwo = undefined
	local pSize = 12
	
	local flagChildren
	
	--close Dialog if already open	
	try(destroyDialog scaleMoveTool) catch()
	
	--Functions
	fn crossPoints_fn pnts = 
	(
		for p in pnts where isValidNode p do
		(
			p.centermarker = off
			p.cross = on
			p.axistripod = off
			p.Box = off
			p.Size = pSize
			p.wirecolor = (color 255 0 0)
		)
	)
	
	fn createPoints_fn sel = 
	(
		pointArray = #()
		for s in sel where isValidNode s do
		(
			p = point name:(uniquename "SMT_Point_")
			s.parent = undefined
			p.pos.controller = s.pos.controller
			p.rotation.controller = s.rotation.controller
			append pointArray p
		)
		crossPoints_fn pointArray
		return pointArray
	)
	
	fn masterPoint_fn pnt picObj = 
	(
	 	case pnt of
		(
			1:(
				pOne = picObj
				
				scaleMoveTool.pic_pOne.text = pOne.name
			  )--end 1
			2:(
				pTwo = picObj
				
				scaleMoveTool.pic_pTwo.text = pTwo.name
			  ) --end 2
		)--end case
		
		crossPoints_fn thePoints
		
		for p in #(pOne, pTwo) where isValidNode p do
		(
			p.centermarker = on
			p.cross = off
			p.axistripod = off
			p.Box = on
			p.size *= 2
		)
		
		if pOne != undefined AND pTwo != undefined do scaleMoveTool.chk_skew.enabled = true
		
	)--end fn
	
	fn restorePosCons_fn arr = 
	(
		for i = 1 to arr.count where isValidNode arr[i][1] do
		(
			curPos = arr[i][1].pos
			arr[i][1].pos.controller = copy arr[i][1].pos.controller[1]
			arr[i][1].pos = curPos
			if isValidNode arr[i][2] do
			(
				arr[i][2].pos.controller = arr[i][1].pos.controller
			)
		)
	)
	
	fn reParentObjects_fn objs objPs = 
	(
		for i = 1 to objs.count do 
		(
			 if isValidNode objs[i] AND isValidNode objPs[i] AND objs[i] != objPs[i] do 
			 	objs[i].parent = objPs[i]
		)
		select (for o in objs where isValidNode o collect o)
	)
	
	fn unLinkChildren_fn objs = 
	(
		childSuperArray = #()
		for o in objs do
		(
		childMiniArray = #()
			if o.children.count > 0 then
			(
				tempArray = (for c in o.children collect c)
				for t in tempArray do
				(
					t.parent = undefined
					append childMiniArray t
				)
			)
			else
			(
				append childMiniArray undefined
			)
			append childSuperArray childMiniArray 
		)
		childSuperArray
	)
	
	fn reLinkChildren_fn objs childs = 
	(
		for i = 1 to objs.count where isValidNode objs[i] and childs[i][1] != undefined do
			for c in childs[i] where isValidNode c do c.parent = objs[i]
	)
	
	fn pointsFilter_fn obj = 
	(
		if classof obj == Point AND obj != pOne AND obj != pTwo then 
			return true
		else
			return false
	)
	
	fn objectsFilter_fn obj = 
	(
		o = obj.baseobject
		try
		(
			if classof o != Particle_View AND classof o != Biped_Object AND classof o != BoneGeometry AND
				classof o != DaylightAssemblyHead AND classof obj.parent != DaylightAssemblyHead AND
				classof obj.children[1] != DaylightAssemblyHead then return true
			else return false
		)
		catch (return false)
	)
	
	fn GroupFilter_fn obj = 
	(
		if (NOT isGroupMember obj) AND (NOT isGroupHead obj) then return true
		else return false
	)
	
	fn linkedFilter_fn obj = 
	(
		if obj.parent == undefined AND obj.children[1] == undefined then return true
		else return false
	)

	fn theObjectsIgnoreGroups_fn = 
	(
		objs = for o in selection where objectsFilter_fn o AND GroupFilter_fn o collect o
		for o in objs do o.pos.controller = position_list()
		return objs
	)

	fn theObjectsExplodeGroups_fn = 
	(
		local objs
		temp = for s in selection where objectsFilter_fn s collect s
		for t in temp where isGroupHead t do explodeGroup t
		
		local ig = scaleMoveTool.chk_ignoreLinked_hir
		if ig.checked == true AND ig.enabled = true then
			objs = (for t in temp where isValidNode t AND linkedFilter_fn t collect t)
		else
			objs = (for t in temp where isValidNode t collect t)
		
		for o in objs do o.pos.controller = position_list()
		return objs
	)
	
	fn skewPointsCallbackFilter o = 
	(
		if o == pOne OR o == pTwo then return true
		else return false
	)
	
	/*fn keepSelFilter_fn = 
	(
		SetSelectFilter (GetNumberSelectFilters())
	)*/
	
	fn skewModeStart_fn = 
	(
		max create mode
		
		--disable the master point picking buttons
		scaleMoveTool.pic_pOne.enabled = false
		scaleMoveTool.pic_pTwo.enabled = false		
		
		--master points position controller variables
		p1 = pOne.pos.controller[1]
		p2 = pTwo.pos.controller[1]
		
		--position multiplier variable
		PosMulti = [1,1,1] 
		for i = 1 to 3 where pOne.pos[i] > pTwo.pos[i] do PosMulti[i] = -1
		
		--create the expression for the script controller
		txtSS = stringstream ""
		
		format "curDist = [ (abs(p1[1] - p2[1])), (abs(p1[2] - p2[2])), (abs(p1[3] - p2[3])) ]\n" to:txtSS
		format "for i = 1 to 3 where curDist[i] == 0 do curDist[i] = 0.00001\n" to:txtSS
		format "newVec = [0,0,0]\n" to:txtSS
		format "for i = 1 to 3 do (\n" to:txtSS
		format "if p2[i] > p1[i] then newVec[i] = p1[i] + origVec[i] * posMulti[i] * curDist[i] / origDist[i]\n" to:txtSS
		format "else newVec[i] = p1[i] - origVec[i] * posMulti[i] * curDist[i] / origDist[i] )\n" to:txtSS
		format "newVec" to:txtSS
		
		txt = txtSS as string
		
		--filter the master points from the selection
		skewPoints = for p in thePoints where p != pOne AND p != pTwo collect p
		
		for p in skewPoints do 
		(
			--add a position_list controller if one does not already exist
			--if classof p.pos.controller != Position_List do	p.pos.controller = position_list()
			
			p.pos.available.controller = position_script()
			p.pos.controller.weight[1] = 0 
			
			--this variable must be assigned AFTER the position_list has been created
			posCon = p.pos.controller
			
			--initialize origDist and origVec variables
			--origDist = [ (abs(p1[1].value - p2[1].value)), (abs(p1[2].value - p2[2].value)), (abs(p1[3].value - p2[3].value)) ]
			origDist = pOne.pos - pTwo.pos
			origVec = posCon[1].value - p1.value
			
			--to ensure that later we dont divide by any 0s we change any 0s to 0.00001
			--this value will be close enough for the needs of this script
			for i = 1 to 3 do
			(
				if origDist[i]  < 0 do origDist[i] *= -1
				if origDist[i] == 0 do origDist[i] = 0.00001
				if  origVec[i] == 0 do  origVec[i] = 0.00001
			)
			
			--add the variables to the script controller
			posCon[2].AddConstant "origVec" origVec
			posCon[2].AddConstant "origDist" origDist
			posCon[2].AddConstant "PosMulti" PosMulti
			posCon[2].AddTarget "p1" p1
			posCon[2].AddTarget "p2" p2
			
			posCon.active = 2 --make the script controller the active controller
			
			posCon[2].SetExpression txt --apply the script to the script controller
		)
		
		--create a new selection set for the master points only
		registerSelectFilterCallback skewPointsCallbackFilter "Skew Points"
		SetSelectFilter (GetNumberSelectFilters())
		
		select #(pOne, pTwo)
		selection.size = 2 * pSize
		
	)
	
	fn skewModeStop_fn = 
	(
		try(unregisterSelectFilterCallback skewPointsCallbackFilter)catch()
		
		--enable the master point picking buttons
		scaleMoveTool.pic_pOne.enabled = true
		scaleMoveTool.pic_pTwo.enabled = true
		
		--remove the script controller from theObjects
		for o in theObjects where isValidNode o do 
		(
			curPos = o.pos
			o.pos.controller.weight[1] = 100
			try(o.pos.controller.delete 2)catch()
			o.pos = curPos
		)
		SetSelectFilter 10 --set selection filter to Points
		select (for p in thePoints where isValidNode p collect p)
	)
	
	fn startSMT_fn = 
	(
		if selection.count > 1 then
		(
			max create mode

			case scaleMoveTool.rdo_groups.state of
			(
				1:	(theObjects = theObjectsIgnoreGroups_fn())
				2:	(theObjects = theObjectsExplodeGroups_fn())
			)
			if theObjects.count > 1 then
			(
				selFilter = getSelectFilter()
				SetSelectFilter 10 --set selection filter to Points
				
				scaleMoveTool.ckb_active.text = ">> Stop Scale Move <<"
										
				theObjectsParent = #()
				for o in theObjects do 
					if o.parent != undefined then append theObjectsParent o.parent else append theObjectsParent o
				
				if scaleMoveTool.chk_children.enabled == true AND scaleMoveTool.chk_children.checked == true then
				(
					theObjectsChildren = unLinkChildren_fn theObjects
					flagChildren = true
				)
				else flagChildren = false
				scaleMoveTool.chk_children.enabled = false
				scaleMoveTool.chk_ignoreLinked_hir.enabled = false
				
				thePoints = createPoints_fn theObjects 
				select thePoints
				
				scaleMoveTool.pic_pOne.enabled = true
				scaleMoveTool.pic_pTwo.enabled = true
				
				scaleMoveTool.timerControl.active = true
				
				max scale
			)
			else --goes with: if theObjects.count > 1
			(
				messageBox "Select 2 or more objects" title:"Error" beep:false
				scaleMoveTool.ckb_active.checked = false
			)
		)
		else --goes with: if selection.count > 1
		(
			messageBox "Select 2 or more objects" title:"Error" beep:false
			scaleMoveTool.ckb_active.checked = false
		)
		
	)
	
	fn stopSMT_fn = 
	(
		max create mode

		scaleMoveTool.ckb_active.text = ">> Start Scale Move <<"
		scaleMoveTool.timerControl.active = false
		
		if scaleMoveTool.chk_skew.checked == true do skewModeStop_fn()
		scaleMoveTool.chk_skew.enabled = false
		scaleMoveTool.chk_skew.checked = false
		
		scaleMoveTool.chk_ignoreLinked_hir.enabled = true
		if scaleMoveTool.chk_ignoreLinked_hir.checked == false do
			scaleMoveTool.chk_children.enabled = true
		
		scaleMoveTool.pic_pOne.text = "Pick Master Point 1"
		scaleMoveTool.pic_pOne.enabled = false
		scaleMoveTool.pic_pTwo.text = "Pick Master Point 2"
		scaleMoveTool.pic_pTwo.enabled = false
		
		pOne = pTwo = undefined --clear the master point variables
		
		delete (for p in thePoints where isValidNode p collect p)
		
		if theObjects != undefined do
		(
			reParentObjects_fn theObjects theObjectsParent
			if flagChildren == true do 
				reLinkChildren_fn theObjects theObjectsChildren
				
		)
		
		try(unregisterSelectFilterCallback skewPointsCallbackFilter)catch()
		--try(callbacks.removeScripts #customDisplayFilterChanged id:#SelFilter)catch()
		try(SetSelectFilter selFilter)catch(SetSelectFilter 1)
		
		for o in theObjects where isValidNode o do o.pos.controller = copy o.pos.controller[1]
		
		gc() --force garbage collect
		
		max select
	)
	
	fn sortLinkedObjs_fn arr = 
	(
		--collect all unlinked objects
		temp = for a in arr where a.children.count == 0 AND a.parent == undefined collect a
		--collect all objects with parents but no children and join with the array 'temp'
		join temp (for a in arr where a.children.count == 0 AND a.parent != undefined collect a)
		--go through all items in temp and append their parents to temp. Should work recursively.
		for t in temp where ((findItem arr t.parent) > 0) do append temp t.parent
		--remove duplicate objects
		for i = temp.count to 1 by -1 where (index = findItem temp temp[i]) > 0 AND
			index != i do deleteItem temp index
		--reverse the array order
		sortedArray = for i = temp.count to 1 by -1 collect temp[i]
		
		sortedArray
	)
	
	fn quickAlign_fn a ignore =
	(
		local sel
		case ignore of
		(
			true: (sel = (for s in selection where objectsFilter_fn s AND linkedFilter_fn s collect s))
			false:(sel = sortLinkedObjs_fn (for s in selection where objectsFilter_fn s collect s))
		)
		
		if sel.count > 1 do
		(	
			minVal = 99999999
			maxVal = -99999999
			
			for s in sel do
			(
				if s.pos[a] < minVal do minVal = s.pos[a]
				if s.pos[a] > maxVal do maxVal = s.pos[a]
			)
			case a of 
			(
				1:	(sel.pos.x = (minVal + maxVal) / 2)
				2:	(sel.pos.y = (minVal + maxVal) / 2)
				3:	(sel.pos.z = (minVal + maxVal) / 2)
			)
		)--end if do
	)
	
	--Rollouts
	rollout scaleMoveTool "ScaleMoveTool"
	(
		checkbutton ckb_active ">> Start Scale Move <<" height:30  align:#center
		timer timerControl interval:800 active:false
		Group "Hierarchy Options"
		(
			checkbox chk_ignoreLinked_hir "Ignore Linked Objects" checked:false offset:[-2,0]
			checkbox chk_children "Don't Affect Children" enabled:true offset:[-2,0]
		)
		Group "Group Options"
		(
			radiobuttons rdo_groups labels:#("Deselect Groups", "Explode Groups")
		)
		Group "Point Skewing"
		(
			pickbutton pic_pOne "Pick Master Point 1" width:108 autodisplay:true enabled:false filter:pointsFilter_fn
			pickbutton pic_pTwo "Pick Master Point 2" width:108 autodisplay:true enabled:false filter:pointsFilter_fn
			checkbox chk_skew "Skewing Mode" enabled:false offset:[-2,0]
		)
		Group "Quick Alignment Tools"
		(
			button btn_x "X" across:3
			button btn_y "Y"
			button btn_z "Z"
			checkbox chk_ignoreLinked_qat "Ignore Linked Objects" checked:true offset:[-2,0]
		)
		
		on scaleMoveTool close do
		(
			timerControl.active = false
			if ckb_active.checked == true do
			(
				undo "StopScaleMove" on
				(
					stopSMT_fn true
				)
			)
		)
		
		on ckb_active changed theState do 
		(
			if theState == true then --button is checked
			(
				undo "StartScaleMove" on
				(
					startSMT_fn()
				)
			)
			else --button is unchecked
			(
				undo "StopScaleMove" on
				(					
					stopSMT_fn()
				)
			)
		)
		
		on chk_ignoreLinked_hir changed theState do
		(
			if theState == true then chk_children.enabled = false
			else chk_children.enabled = true
		)
		
		on timerControl tick do
		(
			if thePoints.count > 0 do
			(
				with animate off
				(
					for p in thePoints where isValidNode p do p.scale = [1,1,1]
				)
			)
		)
		
		on pic_pOne picked obj do
		(
			if obj != undefined do
			(
				masterPoint_fn 1 Obj
			)
		)
		
		on pic_pTwo picked obj do
		(
			 if obj != undefined do
			 (
				masterPoint_fn 2 Obj
			 )
		)
		
		on chk_skew changed theState do
		(
			if theState == true AND pOne != undefined AND pTwo != undefined then
			(
				undo "StartSkewMode" on
				(
					skewModeStart_fn()
				)
				
			)
			else if theState == false AND pOne != undefined AND pTwo != undefined then
			(
				undo "StopSkewMode" on
				(
					skewModeStop_fn()
				)				
			)
			else
			(
				chk_skew.checked = false
				messageBox "Pick Two Master Points First" title:"Master Points are undefined" beep:false
			)
		)
		
		on btn_x pressed do
		(
			if selection.count > 1 do
			(
				undo "QuickAlign_X" on
				(
					quickAlign_fn 1 chk_ignoreLinked_qat.checked
				)
			)
		)
		
		on btn_y pressed do
		(
			if selection.count > 1 do
			(
				undo "QuickAlign_Y" on
				(
					quickAlign_fn 2 chk_ignoreLinked_qat.checked
				)
			)
		)
		
		on btn_z pressed do
		(
			if selection.count > 1 do
			(
				undo "QuickAlign_Z" on
				(
					quickAlign_fn 3 chk_ignoreLinked_qat.checked
				)
			)
		)
	)
--create the tool
createDialog scaleMoveTool pos:[110,102] width:144 --auto height

)


